# sci-sundials

Scilab toolbox giving access to Sundials ODE solvers: CVODE, IDA, ARKODE.

## What has been done and what is to be done 

CVODE and IDA has been interfaced and many common features have been factorized in the OdeManager class. However, ARKODE (various modern explicit, implicit and mixed explicit/implicit Runge-Kutta solvers) does not exist in the old 2.4.0 version of Sundials which is used in Scilab hence won't be interfaced unless an upgrade is done (Sundials is now at version 5.7.0). The support of Sparse Jacobians is also missing for the same reason.

## Features

The CVODE gateway cvode() implements the following features which were missing by the legacy LSODE/LSODA/LSODAR/... ode() gateway :

- [ ] full a posteriori access to solver continuous extension of solution at arbitrary points via a MList output when only one lhs is given:
```
sol = cvode(f, [t0 tf], y0)
t = linspace(t0, tf, 1000)
plot(t, sol(t)) 
```
The `sol` MList fields gathers all information related to the obtained solution (time steps, solution at time steps, events, ...). The solver can be restarted by giving the MLlist as first argument to `cvode` :
```
sol2 = cvode(sol, tx, ...)
```
where `tx` is the time point to which solution has to be extended. The options of the call that yielded `sol` are used and  can be changed as optional named parameters after `tx`.
- [ ] a user-friendly access to solver options via optional named parameters in cvode call, i.e.
```
cvode(f, tspan, y0, h0=0.01, rtol=1e-3)
```
- [ ] a really simpler way to give time span of integration allowing to choose between error driven solver internal time steps and user fixed time steps, i.e.
```
[t,y] = cvode(f, [t0 tf], y0)
[t,y] = cvode(f, [t0 t1 ... tf], y0)
[t,t] = cvode(f, tspan, y0, t0=0)
```
the latter style being the closest to actual `ode()` behavior where solution is by default given at user time steps and `t0` is not necessarily equal to `tspan(1)`, which is the default in the two former calls above.
- [ ] a better and user-friendly specification of events via a variable number of outputs  event function (giving value of event equations, wheter to stop integration for a given event and event direction selection), minimal style beeing a single output. Information about events is also simpler to get:
```
function [eq,term,dir] = evfun(t,y)
    eq = y(1)-1.7;
    term = %f;
    dir  = 1;
end
[t,y,te,ye,ie] = cvode(f, tspan, y0, events = evfun)
sol = cvode(f, tspan, y0, events = evfun)
```
in the latter call information about events is recovered in `sol.te,sol.ye,sol.ie`.
- [ ] support of complex solution with detection of complexity in `y0` or `f(t0,y0)`, e.g.
```
function out = crhs(t,y)
    out = 10*exp(2*%i*%pi*t)*y;
end
[t,y] = cvode(crhs, [0,5], 1)
plot(t,real(y),t,imag(y))
```
- [ ] Support of a callback function called after each successfull step, giving access to current solver statistics and allowing to stop integration (e.g. by a "stop" button on a GUI), e.g.
```
function stop = scicallback(t,y,flag,stats)
    stop = %f
    if flag == "step"
        mprintf("%s : hlast=%g\n", flag, stats.hlast)
    end
end
[t,y] = cvode(f, tspan, y0, intcb=scicallback);
```
- [ ] support of an arbitrary number of dimensions of ode state

## Performance

`cvode()` is roughly two times faster than `ode()` for both fixed methods (Adams and BDF). As `ode()` already did, compiled and dynamically linked C,C++ or Fortran externals are supported by `cvode()`. When using such externals instead of Scilab functions `cvode()` is generally an order of magnitude faster.
