//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#ifndef _ODEMANAGER_HXX_
#define _ODEMANAGER_HXX_

#include <array>
#include "callable.hxx"
#include "runvisitor.hxx"
#include "context.hxx"
#include "function.hxx"
#include "double.hxx"
#include "string.hxx"
#include "list.hxx"
#include "mlist.hxx"
#include "pointer.hxx"
#include "struct.hxx"
#include "cell.hxx"

extern "C"
{
#include "localization.h"
#include "Scierror.h"
#include "sciprint.h"

void C2F(daxpy)(int*, double*, double*, int*, double*, int*);
void C2F(dgemv)(char*, int*, int*, double*, double*, int*, double*, int*, double*, double*, int*);
//void C2F(dgemm)(char*, char*, int*, int*, int*, double*, double*, int*, double*, int*, double*, double*, int*);
double C2F(ddot)(int*, double*, int*, double *, int*);

#define ODEMANAGER_KILLME(p) if (p != NULL) p->killMe(); p=NULL;

/* Sundials includes */
#include <nvector/nvector_serial.h>   /* serial N_Vector types, fcts., and macros */
#include <sundials/sundials_dense.h>  /* prototypes for various DlsMat operations */
#include <sundials/sundials_direct.h> /* definitions of DlsMat and DENSE_ELEM */
#include <sundials/sundials_types.h>  /* definition of type realtype */
#include <sundials/sundials_math.h>
}

typedef void(*dynlibFunPtr)();

class OdeManager
{
    public :

    OdeManager(const std::wstring& callerName, const std::wstring& solverName)
    {
        m_wstrCaller = callerName;
        m_wstrSolver = solverName;
        m_iSizeOfInput[EVENTS] = -1;
        m_pCallFunction.fill(NULL);
        m_pDblConstFunction.fill(NULL);
        m_pEntryPointFunction.fill(NULL);
        m_pEntryPointFunctionWithPar.fill(NULL);
        m_dblOptT0 = NAN;
        m_dblInitialStep = NAN;
        m_iNbEvents = 0;
    }

    ~OdeManager() {
        if (m_N_VectorY != NULL) N_VDestroy_Serial(m_N_VectorY);
        if (m_N_VectorYp != NULL) N_VDestroy_Serial(m_N_VectorYp);
        if (m_N_VectorAtol != NULL) N_VDestroy_Serial(m_N_VectorAtol);
        ODEMANAGER_KILLME(m_pDblNSMatrices);
        for (auto pDbl : m_pDblConstFunction)
        {
            ODEMANAGER_KILLME(pDbl);
        }
    };

    enum functionKind {RHS=0,JACY,RES,JACYYP,EVENTS,INTCB};
    enum solverTaskCode {ODE_NORMAL, ODE_ONE_STEP};
    enum solverReturnCode {ODE_SUCCESS=0, ODE_TSTOP_RETURN, ODE_ROOT_RETURN, 
        ODE_CONV_FAILURE, ODE_TOO_MUCH_WORK, ODE_REPTD_RHSFUNC_ERR, ODE_REP_RES_ERR, ODE_WARNING};

    char pstrArgName[6][16] = {"f","\"jacobian\"","\"residual\"","\"jacobian\"","\"events\"","\"intcb\""};

    std::wstring getCallerName()
    {
        return m_wstrCaller;
    }
    std::wstring getSolverName()
    {
        return m_wstrSolver;
    }

    bool isDAE()
    {
        return m_isDAE;
    }
    
    types::Double *getTOut()
    {
        if (m_pDblTOut == NULL)
        {
            // time at the begining of each internal step of the method
            types::Double *pDblTOut = new types::Double(1,m_dblVecTOut.size());
            std::copy(m_dblVecTOut.begin(), m_dblVecTOut.end(), pDblTOut->get());
            return pDblTOut;
        }
        // user prescribed timesteps (eventually stoppped by event)
        return m_pDblTOut;
    }

    types::Double *getYOut()
    {
        // solution y at user prescribed timesteps or at each internal step of the method
        return m_pDblYOut != NULL ? m_pDblYOut : getArrayFromVectors(m_vecYOut);
    }

    types::Double *getYpOut()
    {
        // solution y' at user prescribed timesteps or at each internal step of the method
        return m_pDblYpOut != NULL ? m_pDblYpOut : getArrayFromVectors(m_vecYpOut);
    }

    types::Double *getIndexEvent()
    {
        types::Double *pDbl = createYOut(m_iNbEvents, m_dblVecIndexEvent.size());
        for (int i = 0; i < m_dblVecIndexEvent.size(); i++)
        {
            std::copy(m_dblVecIndexEvent[i].begin(), m_dblVecIndexEvent[i].end(), pDbl->get() + i*m_iNbEvents);
        }
        return pDbl; 
    }

    types::Double *getTEvent()
    {
        types::Double *pDbl = new types::Double(1,m_dblVecEventTime.size());
        std::copy(m_dblVecEventTime.begin(), m_dblVecEventTime.end(), pDbl->get());
        return pDbl;
    }

    types::Double *getYEvent()
    {
        return getArrayFromVectors(m_dblVecYEvent);;            
    }

    types::Double *getYpEvent()
    {
        return getArrayFromVectors(m_dblVecYpEvent);;            
    }

    types::Double *getY0()
    {
        return m_pDblY0;
    }

    std::vector<double> getStepVector()
    {
        return m_dblVecStep;
    }
    std::vector<double> getCurrTimeVector()
    {
        return m_dblVecCurrTime;
    }
    std::vector<int> getPositive()
    {
        return m_iVecPositive;
    }

    int getNbEvents()
    {
        return m_iNbEvents;
    }

    void setIretCount (int iRetCount)
    {
        m_iRetCount = iRetCount;
    }

    int getNEq()
    {
        return m_iNbEq;
    }

    int getNRealEq()
    {
        return m_iNbRealEq;
    }

    dynlibFunPtr getEntryPointFunction(functionKind  what)
    {
        return m_pEntryPointFunction[what];
    }

    dynlibFunPtr getEntryPointFunctionWithPar(functionKind  what)
    {
        return m_pEntryPointFunctionWithPar[what];
    }

    types::typed_list getParameters(functionKind  what)
    {
        return m_pParameters[what];
    }

    types::Double *getConstantFunction(functionKind  what)
    {
        return m_pDblConstFunction[what];
    }

    types::Double *getNordsieckMatrices()
    {
        return m_pDblNSMatrices;
    }

    std::vector<int> getNordsieckMatricesIndex()
    {
        return m_indexNS;
    }

    void parseMatrices(types::typed_list &in);
    types::Double *parseInitialCondition(types::InternalType *pIn, bool bIsDerivative);
    void parseFunction(types::InternalType *pIn, functionKind what);
    void parseOptions(types::optional_list &opt);
    void computeFunction(double t, double *pdblY, double *pdblDY, functionKind what, double *pdblOut, double dblMult = 0);
    void setupEvents();
    bool intermediateCallback(double t, double *pdblY, double *pdblYp, wchar_t *pwstFlag);
    void createSolutionOutput(types::typed_list &out);
    types::Double *createYOut(int iNbOut, int iSizeTSpan, bool bFlat = false);
    types::Double *getArrayFromVectors(std::vector<std::vector<double>> &m_vecY);
    
    void init();
    void solve();

    // virtual OdeManager methods (later instanciated for CVODE, IDA, ARK, ...)
    virtual void *create() = 0;    
    virtual bool initialize(char *) = 0;    
    virtual bool computeIC(char *) = 0;    
    virtual void errHandler(int error_code, const char *module, const char *function, char *msg) = 0;
    virtual bool setUserData() = 0;    
    virtual bool setInitStep() = 0;
    virtual bool setMinStep() = 0;
    virtual bool setMaxStep() = 0;
    virtual bool setStopTime() = 0;
    virtual bool setTolerances() = 0;
    virtual bool setMaxOrder() = 0;
    virtual bool setEventFunction() = 0;
    virtual bool setSolverAndJacobian(char *errorMsg) = 0;
    virtual bool setHandlerFunction() = 0;
    virtual bool getRootInfo(int *piRootsFound) = 0;
    virtual bool getCurrentTime(double *pdblCurrTime) = 0;
    virtual bool getCurrentStep(double *pdblCurrStep) = 0;
    virtual bool getLastOrder(int *piLastOrder) = 0;
    virtual bool getDky(double dblRefTime, int iOrder, N_Vector Y) = 0;
    virtual solverReturnCode doStep(double dblFinalTime, N_Vector Y, N_Vector Yp, double *pdblTime, solverTaskCode iKind) = 0;
    virtual void saveContinuousExtensionVectors() = 0;
    virtual OdeManager *getPreviousManager() = 0;
    virtual types::Struct *getStats() = 0;
    virtual std::vector<std::wstring> getAvailableMethods() = 0;
    virtual int getMaxMethodOrder(std::wstring wstrMethod) = 0;
    virtual void getInterpVectors(double *pdblNS, int iOrderPlusOne, double dblt0, double dblTUser, double dblStep, double *pdblVect, double *pdblVectd) = 0;

    protected :
    
    std::wstring m_wstrMethod;
    std::wstring m_wstrCaller;
    std::wstring m_wstrSolver;
    
    std::array<dynlibFunPtr,6> m_pEntryPointFunction;
    std::array<dynlibFunPtr,6> m_pEntryPointFunctionWithPar;
    std::array<types::Callable *,6> m_pCallFunction;
    std::array<types::Double *,6> m_pDblConstFunction;
    std::array<char *,6> m_pCallFunctionName;
    std::array<int,6> m_iSizeOfInput;
    std::array<types::typed_list,6> m_pParameters;    
    
    N_Vector m_N_VectorY = NULL;
    N_Vector m_N_VectorYp = NULL;
    N_Vector m_N_VectorAtol = NULL;

    types::Callable *m_odeFun = NULL;
    types::Callable *m_odeJac= NULL;

    types::Double *m_pDblTSpan = NULL;    
    types::Double *m_pDblY0 = NULL;    
    types::Double *m_pDblYp0 = NULL;    
    types::Double *m_pDblTOut = NULL;
    types::Double *m_pDblYOut = NULL;
    types::Double *m_pDblYpOut = NULL;

    std::vector<double> m_dblVecAtol;
    std::vector<double> m_dblVecTOut;
    std::vector<double> m_dblVecStep;
    std::vector<double> m_dblVecCurrTime;
    std::vector<int> m_iVecOrder;
    std::vector<int> m_iVecJacBand;
    std::vector<std::vector<double>> m_vecYOut;
    std::vector<std::vector<double>> m_vecYpOut;
    
    std::vector<int> m_iVecEventIsTerminal;
    std::vector<int> m_iVecEventDirection;

    std::vector<std::vector<double>> m_dblVecYEvent;
    std::vector<std::vector<double>> m_dblVecYpEvent;
    std::vector<std::vector<double>> m_dblVecIndexEvent;
    std::vector<double> m_dblVecEventTime;
    std::vector<int> m_iVecPositive;
    std::vector<int> m_iVecIsAlgebric;
    std::wstring m_wstrCalcIc;

    // continuous formula vectors
    // Nordsieck vectors for CVODE, Phi/Psi vectors for IDA
    std::vector<int> m_indexNS;    
    std::list<std::vector<double>> pDblDerOutList;
    types::Double *m_pDblNSMatrices = NULL;
    
    bool m_odeIsComplex = false;
    bool m_odeIsExtension = false;
    bool m_isDAE = false;
    
    double m_dblT0;
    double m_dblOptT0;
    double m_dblRtol;
    double m_dblAtol;
    double m_dblCurrentStep;
    double m_dblInitialStep;  
    double m_dblMaxStep;  
    double m_dblMinStep;  
    int m_iRetCount;
    int m_iNbEq;
    int m_iNbRealEq;
    int m_iMaxOrder;
    int m_iNbEvents;
    int m_iNRefine;   
};

#endif
