//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "CVODEManager.hxx"
#include "CVODEBridge.hxx"

int CVODEFun(realtype t, N_Vector N_VectorY, N_Vector N_VectorYDot, void *pManager)
{
    CVODEManager *manager = static_cast<CVODEManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    // positiveness constraints
    for (auto i : manager->getPositive())
    {
        if (NV_Ith_S(N_VectorY,i-1) < 0)
        {
            return 1; // return a CVODE recoverable error
        }
    }

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::RHS);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::RHS);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        ((CVODEDynFun)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYDot));
    }
    else if (pFuncWithPar != NULL)
    {
        types::Double *pDblPar = manager->getParameters(OdeManager::RHS)[0]->getAs<types::Double>();
        ((CVODEDynFunWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYDot), pDblPar->get());
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::RHS, NV_DATA_S(N_VectorYDot));
    }
    
    return 0;
}

int CVODEJac(long int iNEq, realtype t, N_Vector N_VectorY, N_Vector N_VectorFy, DlsMat DlsMat_J, void *pManager, N_Vector tmp1, N_Vector tmp2, N_Vector tmp3)
{
    CVODEManager *manager = static_cast<CVODEManager *>(pManager);
    int iNbEq = manager->getNEq();
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::JACY);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::JACY);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        ((CVODEDynJac)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorFy), DlsMat_J->data);
    }
    else if (pFuncWithPar != NULL)
    {
        int iNbRealEq = manager->getNRealEq();
        types::Double *pDblPar = manager->getParameters(OdeManager::RHS)[0]->getAs<types::Double>();
        ((CVODEDynJacWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorFy), DlsMat_J->data, pDblPar->get());
    }
    else if (manager->getConstantFunction(OdeManager::JACY) != NULL)
    {
        // Jacobian is constant
        types::Double *pDbl = manager->getConstantFunction(OdeManager::JACY);
        std::copy(pDbl->get(), pDbl->get()+pDbl->getSize(), DlsMat_J->data);
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::JACY, DlsMat_J->data);        
    }
    return 0;
}

