//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "CVODEManager.hxx"
#include "SUNDIALSBridge.hxx"
#include "CVODEBridge.hxx"
#include "odeparameters.hxx"

#define MSGCV_HNIL_STOP "Internal " MSG_TIME_H " are such that t + h = t on the next step. Singularity likely."


void *CVODEManager::create()
{
    m_cvode_mem = NULL;
    
    if (m_wstrMethod == L"adams")
    {
        m_cvode_mem = CVodeCreate(CV_ADAMS, CV_FUNCTIONAL);
    }
    else if (m_wstrMethod == L"bdf")
    {
        m_cvode_mem = CVodeCreate(CV_BDF, CV_NEWTON);
    }

    m_N_VectorY = N_VNew_Serial(m_iNbRealEq);

    return m_cvode_mem;
}
bool CVODEManager::initialize(char *errorMsg)
{
    if (CVodeInit(m_cvode_mem, CVODEFun, m_dblT0, m_N_VectorY) != CV_SUCCESS)
    {
        sprintf(errorMsg, "CVodeInit error");
        return true;
    }
    return false;
}
bool CVODEManager::computeIC(char *errorMsg)
{
    return false;
}
bool CVODEManager::setUserData()
{
    return CVodeSetUserData(m_cvode_mem, (void *)(this)) != CV_SUCCESS;
}
bool CVODEManager::setInitStep()
{
    return CVodeSetInitStep(m_cvode_mem, m_dblInitialStep) != CV_SUCCESS;
}
bool CVODEManager::setMinStep()
{
    return CVodeSetMinStep(m_cvode_mem, m_dblMinStep) != CV_SUCCESS;
}
bool CVODEManager::setMaxStep()
{
    return CVodeSetMaxStep(m_cvode_mem, m_dblMaxStep) != CV_SUCCESS; 
}
bool CVODEManager::setTolerances()
{
    return CVodeSVtolerances(m_cvode_mem, m_dblRtol, m_N_VectorAtol) != CV_SUCCESS;
}
bool CVODEManager::setMaxOrder()
{
    return CVodeSetMaxOrd(m_cvode_mem, m_iMaxOrder) != CV_SUCCESS;
}
bool CVODEManager::setSolverAndJacobian(char *errorMsg)
{
    if (m_wstrMethod == L"bdf")
    {
        if (CVDense(m_cvode_mem, m_iNbRealEq) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVDense error\n");
            return true;
        }

        if (m_pCallFunction[JACY] != NULL || m_pDblConstFunction[JACY] != NULL || m_pEntryPointFunction[JACY] != NULL)
        {   
            if (CVDlsSetDenseJacFn(m_cvode_mem, CVODEJac) != CV_SUCCESS)
            {
                sprintf(errorMsg,"CVDlsSetDenseJacFn error\n");
                return true;
            }
        }
        else
        {
            if (CVDlsSetDenseJacFn(m_cvode_mem, NULL) != CV_SUCCESS)
            {
                sprintf(errorMsg,"CVDlsSetDenseJacFn error\n");
                return true;
            }                    
        }
    }
    return false;
}

bool CVODEManager::setEventFunction()
{
    if (CVodeRootInit(m_cvode_mem, m_iNbEvents, SUNDIALSEvent) != CV_SUCCESS)
    {
        return true;
    }
    if (m_iVecEventDirection.size() > 0)
    {
        if (CVodeSetRootDirection(m_cvode_mem, m_iVecEventDirection.data()) != CV_SUCCESS)
        {
            return true;
        }
    }
    return false;
}
bool CVODEManager::setHandlerFunction()
{
    return CVodeSetErrHandlerFn(m_cvode_mem, SUNDIALSErrHandler, (void *)(this)) != CV_SUCCESS;
}
bool CVODEManager::setStopTime()
{
    return CVodeSetStopTime(m_cvode_mem, m_pDblTSpan->get(m_pDblTSpan->getSize()-1)) != CV_SUCCESS;
}
bool CVODEManager::getLastOrder(int *piLastOrder)
{
    return CVodeGetLastOrder(m_cvode_mem, piLastOrder) != CV_SUCCESS;
}
bool CVODEManager::getCurrentTime(double *pdblCurrTime)
{
    return CVodeGetCurrentTime(m_cvode_mem, pdblCurrTime) != CV_SUCCESS;
}
bool CVODEManager::getCurrentStep(double *pdblCurrStep)
{
    return CVodeGetCurrentStep(m_cvode_mem, pdblCurrStep) != CV_SUCCESS;
}

void CVODEManager::errHandler(int error_code, const char *module, const char *function, char *msg)
{
    char errorMsg[256];
    if (msg == NULL)
    {
        char cvodeMsg[256] = "";
        double dblTime, dblStep;
        CVodeGetCurrentTime(m_cvode_mem, &dblTime);
        CVodeGetLastStep(m_cvode_mem, &dblStep);
        /* Depending on flag, print error message and return error flag */
        switch (error_code)
        {
            case ODE_CONV_FAILURE:
                sprintf(cvodeMsg, MSGCV_CONV_FAILS, dblTime, dblStep);
                break;
            case ODE_TOO_MUCH_WORK:
                sprintf(cvodeMsg, MSGCV_MAX_STEPS, dblTime);
                break;
            case ODE_REPTD_RHSFUNC_ERR:
                sprintf(cvodeMsg, MSGCV_RHSFUNC_REPTD, dblTime);
                break;          
            case ODE_WARNING:
                sprintf(cvodeMsg, MSGCV_HNIL_STOP, dblTime, dblStep);
                break;
            default:
                sprintf(cvodeMsg,"%s at t=%g\n", CVodeGetReturnFlagName((long int) error_code), dblTime);
                break; 
        }        
        sprintf(errorMsg, _("%ls: %s\n"), getCallerName().c_str(), cvodeMsg);
    }
    else
    {
        sprintf(errorMsg, _("%ls: %s \n"), getCallerName().c_str(), msg);
    }        
    if (error_code != ODE_WARNING)
    {
        throw ast::InternalError(errorMsg);        
    }
    else
    {
        sciprint("\n%s",errorMsg);
    }
}

bool CVODEManager::getDky(double dblRefTime, int iOrder, N_Vector Y)
{
    return CVodeGetDky(m_cvode_mem, dblRefTime, iOrder, Y) != CV_SUCCESS;   
}

// Stepper function. Note: CVODE does not return Yp, but prototype is imposed by Odemanager class
// to include IDA stepper prototype, which yields Yp.

OdeManager::solverReturnCode CVODEManager::doStep(double dblFinalTime, N_Vector Y, N_Vector Yp, double *pdblTime, solverTaskCode iKind)
{
    std::map<solverTaskCode, int> toCVODETask = {{ODE_ONE_STEP, CV_ONE_STEP}, {ODE_NORMAL, CV_NORMAL}};
    std::map<int, solverReturnCode> toODEReturn = {
    {CV_SUCCESS, ODE_SUCCESS}, 
    {CV_TSTOP_RETURN, ODE_TSTOP_RETURN}, 
    {CV_ROOT_RETURN, ODE_ROOT_RETURN},
    {CV_CONV_FAILURE, ODE_CONV_FAILURE},
    {CV_TOO_MUCH_WORK, ODE_TOO_MUCH_WORK},
    {CV_REPTD_RHSFUNC_ERR, ODE_REPTD_RHSFUNC_ERR},
    {CV_WARNING, ODE_WARNING}
    };
    
    int iFlag = CVode(m_cvode_mem, dblFinalTime, Y, pdblTime, toCVODETask[iKind]);    
    
    return toODEReturn[iFlag];
}

void CVODEManager::saveContinuousExtensionVectors()
{
    CVodeMem cv_mem = (CVodeMem) m_cvode_mem;
    // Solution structure output
    // Store current ordrer and Nordsieck History Array
    // N_Vector cv_zn[L_MAX];   Nordsieck array, of size N x (q+1).
    // zn[j] is a vector of length N (j=0,...,q)
    // zn[j] = [1/factorial(j)] * h^j *(jth derivative of the interpolating polynomial)             
    // indexNS is the cumulative index later used for solution interpolation
    // indexNS[i] is the column index in whole matrix where NS matrix of timestep i starts.
    m_indexNS.push_back(m_indexNS.back()+m_iVecOrder.back()+1);

    for (int i=0; i<m_iVecOrder.back()+1; i++)
    {
        std::vector<double> vdblNordsieckVector (NV_DATA_S(cv_mem->cv_zn[i]), NV_DATA_S(cv_mem->cv_zn[i]) + m_iNbRealEq);
        pDblDerOutList.push_back(vdblNordsieckVector);           
    }
}

void CVODEManager::getInterpVectors(double *pdblNS, int iOrderPlusOne, double dblt0, double dblTUser, double dblStep, double *pdblVect, double *pdblVectd)
{
    // pdblNS matrix of Nordsieck vectors not used here (unlike IDA)
    double dblS = (dblTUser-dblt0)/dblStep;
    pdblVect[0] = 1;
    pdblVectd[0] = 0;    
    for (int j = 1; j < iOrderPlusOne; j++)
    {
        pdblVectd[j] = pdblVect[j-1]*j/dblStep;
        pdblVect[j] = pdblVect[j-1]*dblS;
    } 
}

bool CVODEManager::getRootInfo(int *piRootsFound)
{
    return CVodeGetRootInfo(m_cvode_mem, piRootsFound) != CV_SUCCESS;
}

types::Struct *CVODEManager::getStats()
{
    double dblStat[6] = {0.0,0.0,0.0,0.0,0.0,0.0};
    int qlast;
    int qcur;

    std::wstring fieldNames[14] = {L"nsteps", L"nfevals", L"nfevalsFD", L"njacevals", L"neventevals",
    L"nlinsolve", L"nrejsteps", L"nnliters", L"nnlcvfails", L"order", L"h0", L"hlast",
    L"hcur", L"tcur"};
    
    if (m_pCallFunction[JACY] != NULL)
    {
        m_incStat[2] = 0;
    }

    CVodeGetNonlinSolvStats(m_cvode_mem, m_incStat+7, m_incStat+8);
    CVodeGetIntegratorStats(m_cvode_mem, m_incStat, m_incStat+1, m_incStat+5, m_incStat+6, &qlast, &qcur, dblStat, dblStat+1, dblStat+2, dblStat+3);
    CVodeGetNumGEvals(m_cvode_mem, m_incStat+4);
    CVodeGetNumLinSolvSetups(m_cvode_mem,m_incStat+5); // CVodeGetIntegratorStats seems broken for nlinsolve(m_incStat+5), hence we do a direct call

    if (m_wstrMethod == L"bdf")
    {
        CVDlsGetNumRhsEvals(m_cvode_mem, m_incStat+2);
        CVDlsGetNumJacEvals(m_cvode_mem, m_incStat+3);        
    }
    
    // dblStat[0] = (double)qlast;
    // dblStat[1] = (double)qcur;

    // if extending a previous solution, update incremental stats only
    if (m_prevManager != NULL)
    {
        for (int i=0; i<9; i++)
        {
           m_incStat[i] += m_prevManager-> m_incStat[i]; 
        }
    }
    
    types::Struct *pSt = new types::Struct(1,1);
    for (int i=0; i<9; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(),new types::Double((double)m_incStat[i]));        
    }

    // order of method for each step
    types::Double *pDblOrder = new types::Double(1,m_iVecOrder.size());
    std::copy(m_iVecOrder.begin(), m_iVecOrder.end(), pDblOrder->get());
    pSt->addField(fieldNames[9].c_str());        
    pSt->get(0)->set(fieldNames[9].c_str(), pDblOrder);        
    
    for (int i=10; i<14; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(), new types::Double(dblStat[i-10]));        
    }
    return pSt;
}