//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "IDAManager.hxx"
#include "SUNDIALSBridge.hxx"
#include "IDABridge.hxx"
#include "odeparameters.hxx"

#define MSGCV_HNIL_STOP "Internal " MSG_TIME_H " are such that t + h = t on the next step. Singularity likely."

void *IDAManager::create()
{    
    m_ida_mem = IDACreate();
    m_N_VectorY = N_VNew_Serial(m_iNbRealEq);
    m_N_VectorYp = N_VNew_Serial(m_iNbRealEq);

    return m_ida_mem;
}

bool IDAManager::initialize(char *errorMsg)
{
    if (IDAInit(m_ida_mem, IDARes, m_dblT0, m_N_VectorY, m_N_VectorYp) != IDA_SUCCESS)
    {
        sprintf(errorMsg,"IDAInit error.");
    }
    return false;
}

bool IDAManager::computeIC(char *errorMsg)
{
    // compute initial condition, if applicable
    if (m_wstrCalcIc == L"y0yp0")
    {
        // Compute yp0 and algebric components of y0 given differential components of y0
        N_Vector id = N_VNew_Serial(m_iNbRealEq);
        std::fill(NV_DATA_S(id), NV_DATA_S(id)+m_iNbRealEq, 1);
        for (auto index : m_iVecIsAlgebric)
        {
            NV_DATA_S(id)[index-1] = 0; // 0 means algebric state   
            if (m_odeIsComplex)
            {
                NV_DATA_S(id)[index-1+m_iNbEq] = 0; // 0 means algebric state                   
            }
        }
        if  (IDASetId(m_ida_mem, id) != IDA_SUCCESS)
        {
            sprintf(errorMsg,"IDASetId error\n");
            return true;                   
        }

        long int iFlag = IDACalcIC(m_ida_mem, IDA_YA_YDP_INIT, m_pDblTSpan->get(m_pDblTSpan->getSize()-1));

        if (iFlag != IDA_SUCCESS)
        {
            sprintf(errorMsg,"IDACalcIC error : %s\n", IDAGetReturnFlagName(iFlag));
            return true;
        }        
    } 
    else if (m_wstrCalcIc == L"y0")
    {
        // Compute y0 given yp0
        long int iFlag = IDACalcIC(m_ida_mem, IDA_Y_INIT, m_pDblTSpan->get(m_pDblTSpan->getSize()-1));
        if (iFlag != IDA_SUCCESS)
        {
            sprintf(errorMsg,"IDACalcIC error : %s\n", IDAGetReturnFlagName(iFlag));
            return true;                
        }        
    }
    if (IDAGetConsistentIC(m_ida_mem, m_N_VectorY, m_N_VectorYp) != IDA_SUCCESS)
    {
        sprintf(errorMsg,"IDAGetConsistentIC error\n");
        return true;                        
    }
    return false;
}
bool IDAManager::setUserData()
{
    return IDASetUserData(m_ida_mem, (void *)(this)) != IDA_SUCCESS;
}
bool IDAManager::setInitStep()
{
    return IDASetInitStep(m_ida_mem, m_dblInitialStep) != IDA_SUCCESS;
}
bool IDAManager::setMinStep()
{
    return false;
}
bool IDAManager::setMaxStep()
{
    return IDASetMaxStep(m_ida_mem, m_dblMaxStep) != IDA_SUCCESS; 
}
bool IDAManager::setTolerances()
{
    return IDASVtolerances(m_ida_mem, m_dblRtol, m_N_VectorAtol) != IDA_SUCCESS;
}
bool IDAManager::setMaxOrder()
{
    return IDASetMaxOrd(m_ida_mem, m_iMaxOrder) != IDA_SUCCESS;
}
bool IDAManager::setSolverAndJacobian(char *errorMsg)
{
    if (IDADense(m_ida_mem, m_iNbRealEq) != IDA_SUCCESS)
    {
         sprintf(errorMsg,"IDADense error\n");
         return true;         
    }
    if (m_pCallFunction[JACYYP] != NULL || m_pDblConstFunction[JACYYP] != NULL || m_pEntryPointFunction[JACYYP] != NULL)
    {   
        if (IDADlsSetDenseJacFn(m_ida_mem, IDAJac) != IDA_SUCCESS)
        {
            sprintf(errorMsg,"IDADlsSetDenseJacFn error\n");
            return true;
        }
    }
    return false;
}

bool IDAManager::setEventFunction()
{
    if (IDARootInit(m_ida_mem, m_iNbEvents, SUNDIALSImplEvent) != IDA_SUCCESS)
    {
        return true;
    }
    if (m_iVecEventDirection.size() > 0)
    {
        if (IDASetRootDirection(m_ida_mem, m_iVecEventDirection.data()) != IDA_SUCCESS)
        {
            return true;
        }
    }
    return false;
}
bool IDAManager::setHandlerFunction()
{
    return IDASetErrHandlerFn(m_ida_mem, SUNDIALSErrHandler, (void *)(this)) != IDA_SUCCESS;
}
bool IDAManager::setStopTime()
{
    return IDASetStopTime(m_ida_mem, m_pDblTSpan->get(m_pDblTSpan->getSize()-1)) != IDA_SUCCESS;
}
bool IDAManager::getLastOrder(int *piLastOrder)
{
    return IDAGetLastOrder(m_ida_mem, piLastOrder) != IDA_SUCCESS;
}
bool IDAManager::getCurrentTime(double *pdblCurrTime)
{
    return IDAGetCurrentTime(m_ida_mem, pdblCurrTime) != IDA_SUCCESS;
}
bool IDAManager::getCurrentStep(double *pdblCurrStep)
{
    return IDAGetCurrentStep(m_ida_mem, pdblCurrStep) != IDA_SUCCESS;
}

void IDAManager::errHandler(int error_code, const char *module, const char *function, char *msg)
{
    char errorMsg[256];
    if (msg == NULL)
    {
        char idaMsg[256] = "";
        double dblTime, dblStep;
        IDAGetCurrentTime(m_ida_mem, &dblTime);
        IDAGetLastStep(m_ida_mem, &dblStep);
        /* Depending on flag, print error message and return error flag */
        switch (error_code)
        {
            case ODE_CONV_FAILURE:
                sprintf(idaMsg, MSG_CONV_FAILS, dblTime, dblStep);
                break;
            case ODE_TOO_MUCH_WORK:
                sprintf(idaMsg, MSG_MAX_STEPS, dblTime);
                break;
            case ODE_REP_RES_ERR:
                sprintf(idaMsg, MSG_REP_RES_ERR, dblTime);
                break;          
            case ODE_WARNING:
                sprintf(idaMsg, MSGCV_HNIL_STOP, dblTime, dblStep);
                break;
            default:
                sprintf(idaMsg,"%s at t=%g\n", IDAGetReturnFlagName((long int) error_code), dblTime);
                break; 
        }        
        sprintf(errorMsg, _("%ls: %s\n"), getCallerName().c_str(), idaMsg);
    }
    else
    {
        sprintf(errorMsg, _("%ls: %s \n"), getCallerName().c_str(), msg);
    }        
    if (error_code != ODE_WARNING)
    {
        throw ast::InternalError(errorMsg);        
    }
    else
    {
        sciprint("\n%s",errorMsg);
    }
}

bool IDAManager::getDky(double dblRefTime, int iOrder, N_Vector Y)
{
    return IDAGetDky(m_ida_mem, dblRefTime, iOrder, Y) != IDA_SUCCESS;   
}

// Stepper function

OdeManager::solverReturnCode IDAManager::doStep(double dblFinalTime, N_Vector Y, N_Vector Yp, double *pdblTime, solverTaskCode iKind)
{
    std::map<solverTaskCode, int> toIDATask = {{ODE_ONE_STEP, IDA_ONE_STEP}, {ODE_NORMAL, IDA_NORMAL}};
    std::map<int, solverReturnCode> toODEReturn = {
    {IDA_SUCCESS, ODE_SUCCESS}, 
    {IDA_TSTOP_RETURN, ODE_TSTOP_RETURN}, 
    {IDA_ROOT_RETURN, ODE_ROOT_RETURN},
    {IDA_CONV_FAIL, ODE_CONV_FAILURE},
    {IDA_TOO_MUCH_WORK, ODE_TOO_MUCH_WORK},
    {IDA_REP_RES_ERR, ODE_REP_RES_ERR},
    {IDA_WARNING, ODE_WARNING}
    };    
    
    int iFlag = IDASolve(m_ida_mem, dblFinalTime, pdblTime, Y, Yp, IDA_ONE_STEP);    
    
    return toODEReturn[iFlag];
}

void IDAManager::saveContinuousExtensionVectors()
{
    IDAMem ida_mem = (IDAMem) m_ida_mem; 
    m_indexNS.push_back(m_indexNS.back()+m_iVecOrder.back()+1);

    for (int i=0; i<m_iVecOrder.back()+1; i++)
    {
        std::vector<double> vdblPhiVector (NV_DATA_S(ida_mem->ida_phi[i]), NV_DATA_S(ida_mem->ida_phi[i]) + m_iNbRealEq);
        // add scalar psi[i] at the end of vector phi[i]
        vdblPhiVector.push_back(ida_mem->ida_psi[i]);
        pDblDerOutList.push_back(vdblPhiVector);
    }
}

void IDAManager::getInterpVectors(double *pdblNS, int iOrderPlusOne, double dblt0, double dblTUser, double dblStep, double *pdblVect, double *pdblVectd)
{
    double dblGamma = 0;
    double dblDelta = dblTUser-dblt0;
    double *pdblPsi = pdblNS + m_iNbRealEq; // psi[0] is the last value of first column (above lines are components of phi0)
    double psi_jm1 = *pdblPsi; // psi[0]
    double psi_j = 0; // psi0

    // code is adapted from IDAGetSolution(() in SUNDIALS ida.c
    dblGamma = dblDelta/psi_jm1;
    pdblVect[0] = 1;
    pdblVectd[0] = 0;
    for (int j = 1; j < iOrderPlusOne; j++)
    {
        pdblVectd[j] = pdblVectd[j-1]*dblGamma + pdblVect[j-1]/psi_jm1;
        pdblVect[j] = pdblVect[j-1]*dblGamma;
        pdblPsi += m_iNbRealEq+1; // NS matrix has m_nbRealEq+1 lines, psi[j] is last element of column j
        psi_j = *pdblPsi;
        dblGamma = (dblDelta + psi_jm1) / psi_j;
        psi_jm1 = psi_j;
    }
}

bool IDAManager::getRootInfo(int *piRootsFound)
{
    return IDAGetRootInfo(m_ida_mem, piRootsFound) != IDA_SUCCESS;
}

types::Struct *IDAManager::getStats()
{
    double dblStat[6] = {0.0,0.0,0.0,0.0,0.0,0.0};
    int qlast;
    int qcur;

    std::wstring fieldNames[14] = {L"nsteps", L"nfevals", L"nfevalsFD", L"njacevals", L"neventevals",
    L"nlinsolve", L"nrejsteps", L"nnliters", L"nnlcvfails", L"order", L"h0", L"hlast",
    L"hcur", L"tcur"};
    
    if (m_pCallFunction[JACY] != NULL)
    {
        m_incStat[2] = 0;
    }

    IDAGetNonlinSolvStats(m_ida_mem, m_incStat+7, m_incStat+8);
    IDAGetIntegratorStats(m_ida_mem, m_incStat, m_incStat+1, m_incStat+5, m_incStat+6, &qlast, &qcur, dblStat, dblStat+1, dblStat+2, dblStat+3);
    IDAGetNumGEvals(m_ida_mem, m_incStat+4);
    IDAGetNumLinSolvSetups(m_ida_mem,m_incStat+5); // IDAGetIntegratorStats seems broken for nlinsolve(m_incStat+5), hence we do a direct call
    IDADlsGetNumResEvals(m_ida_mem, m_incStat+2);
    IDADlsGetNumJacEvals(m_ida_mem, m_incStat+3);        
    
    // dblStat[0] = (double)qlast;
    // dblStat[1] = (double)qcur;

    // if extending a previous solution, update incremental stats only
    if (m_prevManager != NULL)
    {
        for (int i=0; i<9; i++)
        {
           m_incStat[i] += m_prevManager-> m_incStat[i]; 
        }
    }
    
    types::Struct *pSt = new types::Struct(1,1);
    for (int i=0; i<9; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(),new types::Double((double)m_incStat[i]));        
    }

    // order of method for each step
    types::Double *pDblOrder = new types::Double(1,m_iVecOrder.size());
    std::copy(m_iVecOrder.begin(), m_iVecOrder.end(), pDblOrder->get());
    pSt->addField(fieldNames[9].c_str());        
    pSt->get(0)->set(fieldNames[9].c_str(), pDblOrder);        
    
    for (int i=10; i<14; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(), new types::Double(dblStat[i-10]));        
    }
    return pSt;
}