//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "SUNDIALSBridge.hxx"
#include "callable.hxx"
#include "runvisitor.hxx"
#include "context.hxx"

void OdeManager::computeFunction(double t, double *pdblY, double *pdblYp, functionKind what, double *pdblOut, double dblMult)
{
    types::typed_list in;
    types::typed_list out;
    types::optional_list opt;
    types::Double* pDblT = NULL;
    types::Double* pDblY = NULL;
    types::Double* pDblYp = NULL;
    types::Double* pDblOut[2] = {NULL,NULL};
    types::Callable *pCall = NULL;
    int iRetCount = 1;

    char errorMsg[256] = "";

    pCall =  m_pCallFunction[what];
    if (pCall == NULL)
    {
        sprintf(errorMsg, "Error, m_pCallFunction[%d] is NULL !",(int)what);
        throw ast::InternalError(errorMsg);        
    }

    in.push_back(new types::Double(t));

    // keep native dimensions of Y0
    pDblY = m_pDblY0->clone();
    if (pdblY != NULL) // can be the case when testing if ode is complex
    {
        std::copy(pdblY, pdblY+m_iNbEq, pDblY->get());
        if (m_odeIsComplex)
        {
            std::copy(pdblY+m_iNbEq, pdblY+2*m_iNbEq, pDblY->getImg());
        }       
    }
    in.push_back(pDblY);

    if (m_isDAE)
    {
        pDblYp = m_pDblYp0->clone();
        if (pdblYp != NULL)
        {
            memcpy(pDblYp->get(), pdblYp, sizeof(double)*m_iNbEq);
        }        
        in.push_back(pDblYp);
    }

    // optional user input parameters 
    for (auto pIn : m_pParameters[what])
    {
        in.push_back(pIn);
    }

    for (auto pIn : in)
    {
        pIn->IncreaseRef();
    }
    
    // if called with what == JACYYP, user function yields [dRes/dY , dRes/dYp]
    if (what == JACYYP)
    {
        iRetCount = 2;
    }
    
    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        pCall->invoke(in, opt, iRetCount, out, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
        for (auto pIn :  m_pParameters[what])
        {
            pIn->DecreaseRef();
        }
        if (ConfigVariable::getLastErrorFunction() == L"")
        {
            // necessary for errors such as "Wrong number of input arguments"
            std::wostringstream ostr;
            ostr << m_pCallFunctionName[what] << L": " <<  ie.GetErrorMessage();
            throw ast::InternalError(ostr.str());          
        }
        else
        {
            sprintf(errorMsg, wide_string_to_UTF8(ie.GetErrorMessage().c_str()));
            throw ast::InternalError(errorMsg);          
        }
    }
    for (auto pOut : out) // safety IncreaseRef
    {
        pOut->IncreaseRef();
    }

    for (auto pIn : in)
    {
        pIn->DecreaseRef();
        if (pIn->isDeletable())
        {
            delete pIn;
        }
    }

    if (out.size() != iRetCount)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), m_pCallFunctionName[what], iRetCount);
        throw ast::InternalError(errorMsg);          
    }

    // Double matrices expected

    for (int i=0; i < iRetCount; i++)
    {
        if (out[i]->isDouble() == false)
        {
            sprintf(errorMsg, _("%s: Wrong type for output argument #%d: double matrix expected.\n"), m_pCallFunctionName[what], i+1);
            throw ast::InternalError(errorMsg);          
        }

        pDblOut[i] = out[i]->getAs<types::Double>();

        m_odeIsComplex |= pDblOut[i]->isComplex();

        if (pDblOut[i]->getSize() != m_iSizeOfInput[what])
        {
            sprintf(errorMsg, _("%s: Wrong size for output argument #%d: expecting %d.\n"), m_pCallFunctionName[what], i, m_iSizeOfInput[what]);
            throw ast::InternalError(errorMsg);          
        }                
    }
    
    if (pdblOut != NULL) // pdblOut == NULL when testing if ode is complex
    {
        int iSize = m_iSizeOfInput[what];
        std::copy(pDblOut[0]->get(), pDblOut[0]->get()+iSize, pdblOut);
        if (m_odeIsComplex && pDblOut[0]->isComplex())
        {
            std::copy(pDblOut[0]->getImg(), pDblOut[0]->getImg()+iSize, pdblOut+iSize);
        }
        // if called with what == JACYYP, compute dRes/dY+dblmult*dRes/dYp
        if (iRetCount == 2)
        {
            int iOne = 1;
            C2F(daxpy)(&iSize, &dblMult, pDblOut[1]->get(), &iOne, pdblOut, &iOne);
            if (m_odeIsComplex && pDblOut[1]->isComplex())
            {
                C2F(daxpy)(&iSize, &dblMult, pDblOut[1]->getImg(), &iOne, pdblOut+iSize, &iOne);                
            }      
        }
    }

    for (int i=0; i < iRetCount; i++)
    {
        pDblOut[i]->killMe();        
    }
}

bool OdeManager::intermediateCallback(double t, double *pdblY, double *pdblYp, wchar_t *pwstFlag)
{
    types::typed_list in;
    types::typed_list out;
    types::optional_list opt;
    types::Double* pDblT = NULL;
    types::Double* pDblY = NULL;
    types::Double* pDblYp = NULL;
    types::String* pStr = NULL;
    types::Callable *pCall = NULL;
    bool bTerm = false;
    char errorMsg[256] = "";

    if (m_pCallFunction[INTCB] == NULL)
    {
        return bTerm;
    }

    in.push_back(new types::Double(t));

    // keep native dimensions of Y0
    pDblY = m_pDblY0->clone();
    if (pdblY != NULL)
    {
        std::copy(pdblY, pdblY+m_iNbEq, pDblY->get());
        if (m_odeIsComplex)
        {
            std::copy(pdblY+m_iNbEq, pdblY+2*m_iNbEq, pDblY->getImg());
        }       
    }
    in.push_back(pDblY);

    if (m_isDAE)
    {
        pDblYp = m_pDblYp0->clone();
        if (pdblYp != NULL)
        {
            memcpy(pDblYp->get(), pdblYp, sizeof(double)*m_iNbEq);
        }        
        in.push_back(pDblYp);
    }

    in.push_back(new types::String(pwstFlag));                

    in.push_back(getStats());

    // optional user input parameters 
    for (auto pIn : m_pParameters[INTCB])
    {
        in.push_back(pIn);
    }

    for (auto pIn : in)
    {
        pIn->IncreaseRef();
    }

    pCall =  m_pCallFunction[INTCB];

    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        pCall->invoke(in, opt, 1, out, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
        for (auto pIn :  m_pParameters[INTCB])
        {
            pIn->DecreaseRef();
        }
        if (ConfigVariable::getLastErrorFunction() == L"")
        {
            // necessary for errors such as "Wrong number of input arguments"
            std::wostringstream ostr;
            ostr << m_pCallFunctionName[INTCB] << L": " <<  ie.GetErrorMessage();
            throw ast::InternalError(ostr.str());          
        }
        else
        {
            sprintf(errorMsg, wide_string_to_UTF8(ie.GetErrorMessage().c_str()));
            throw ast::InternalError(errorMsg);          
        }
    }
    for (auto pOut : out) // safety IncreaseRef
    {
        pOut->IncreaseRef();
    }

    for (auto pIn : in)
    {
        pIn->DecreaseRef();
        if (pIn->isDeletable())
        {
            delete pIn;
        }
    }

    if (out.size() > 1)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), m_pCallFunctionName[INTCB], 1);
        throw ast::InternalError(errorMsg);          
    }
    else if (out.size() > 0)
    {
        // Boolean matrix expected
        if (out[0]->isBool() == false || out[0]->getAs<types::Bool>()->getSize() != 1)
        {
            sprintf(errorMsg, _("%s: Wrong type for output argument #%d: scalar boolean expected.\n"), m_pCallFunctionName[INTCB], 1);
            throw ast::InternalError(errorMsg);          
        }
    
        bTerm = out[0]->getAs<types::Bool>()->get(0);
        out[0]->DecreaseRef();
        out[0]->killMe();
    }

    return bTerm;
}

void OdeManager::setupEvents()
{
    types::typed_list in;
    types::typed_list out;
    types::optional_list opt;
    types::Double* pDblT = NULL;
    types::Double* pDblY = NULL;
    types::Double* pDblYp = NULL;
    types::Double* pDblOut = NULL;
    types::Callable *pCall = NULL;

    int iRetCount = 0;
  
    char errorMsg[256] = "";
     
    pCall =  m_pCallFunction[EVENTS];
    if (pCall == NULL)
    {
        dynlibFunPtr pFunc = m_pEntryPointFunction[EVENTS];
        dynlibFunPtr pFuncWithPar = m_pEntryPointFunctionWithPar[EVENTS];
        if (m_isDAE)
        {
            if (pFunc != NULL)
            {
                // First call to get number of events in m_iNbEvents
                m_iNbEvents = 0;
                ((SUNDIALSDynImplEvent)pFunc)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), m_pDblYp0->get(), &m_iNbEvents, NULL, NULL, NULL);
                m_iVecEventIsTerminal.resize(m_iNbEvents);
                m_iVecEventDirection.resize(m_iNbEvents);
                // Second call to get direction and termination
                ((SUNDIALSDynImplEvent)pFunc)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), m_pDblYp0->get(), &m_iNbEvents, NULL, m_iVecEventIsTerminal.data(),  m_iVecEventDirection.data());            
            }
            else if (pFuncWithPar != NULL)
            {
                // First call to get number of events in m_iNbEvents
                types::Double *pDblPar = getParameters(EVENTS)[0]->getAs<types::Double>();
                m_iNbEvents = 0;
                ((SUNDIALSDynImplEventWithPar)pFuncWithPar)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), m_pDblYp0->get(), &m_iNbEvents, NULL, NULL, NULL, pDblPar->get());
                m_iVecEventIsTerminal.resize(m_iNbEvents);
                m_iVecEventDirection.resize(m_iNbEvents);
                // Second call to get direction and termination
                ((SUNDIALSDynImplEventWithPar)pFuncWithPar)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), m_pDblYp0->get(), &m_iNbEvents, NULL, m_iVecEventIsTerminal.data(),  m_iVecEventDirection.data(), pDblPar->get());  
            }
        }
        else
        {
            if (pFunc != NULL)
            {
                // First call to get number of events in m_iNbEvents
                m_iNbEvents = 0;
                ((SUNDIALSDynEvent)pFunc)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), &m_iNbEvents, NULL, NULL, NULL);
                m_iVecEventIsTerminal.resize(m_iNbEvents);
                m_iVecEventDirection.resize(m_iNbEvents);
                // Second call to get direction and termination
                ((SUNDIALSDynEvent)pFunc)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), &m_iNbEvents, NULL, m_iVecEventIsTerminal.data(),  m_iVecEventDirection.data());            
            }
            else if (pFuncWithPar != NULL)
            {
                // First call to get number of events in m_iNbEvents
                types::Double *pDblPar = getParameters(EVENTS)[0]->getAs<types::Double>();
                m_iNbEvents = 0;
                ((SUNDIALSDynEventWithPar)pFuncWithPar)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), &m_iNbEvents, NULL, NULL, NULL, pDblPar->get());
                m_iVecEventIsTerminal.resize(m_iNbEvents);
                m_iVecEventDirection.resize(m_iNbEvents);
                // Second call to get direction and termination
                ((SUNDIALSDynEventWithPar)pFuncWithPar)(&m_iNbRealEq, &m_dblT0, m_pDblY0->get(), &m_iNbEvents, NULL, m_iVecEventIsTerminal.data(),  m_iVecEventDirection.data(), pDblPar->get());  
            }            
        }
        return;
    }
     
    in.push_back(new types::Double(m_pDblTSpan->get(0)));

    // keep native dimensions of Y0
    pDblY = m_pDblY0->clone();
    in.push_back(pDblY);

    if (m_isDAE)
    {
        pDblYp = m_pDblYp0->clone();
        in.push_back(pDblYp);
    }

    // optional user input parameters 
    for (auto pIn : m_pParameters[EVENTS])
    {
        in.push_back(pIn);
    }
    
    for (auto pIn : in)
    {
        pIn->IncreaseRef();
    }
    
    iRetCount = 3;
    
    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        pCall->invoke(in, opt, iRetCount, out, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
        for (auto pIn :  m_pParameters[EVENTS])
        {
            pIn->DecreaseRef();
        }
        if (ConfigVariable::getLastErrorFunction() == L"")
        {
            // necessary for errors such as "Wrong number of input arguments"
            std::wostringstream ostr;
            ostr << m_pCallFunctionName[EVENTS] << L": " <<  ie.GetErrorMessage();
            throw ast::InternalError(ostr.str());          
        }
        else
        {
            sprintf(errorMsg, wide_string_to_UTF8(ie.GetErrorMessage().c_str()));
            throw ast::InternalError(errorMsg);          
        }
    }
    for (auto pOut : out) // safety IncreaseRef
    {
        pOut->IncreaseRef();
    }

    for (auto pIn : in)
    {
        pIn->DecreaseRef();
        if (pIn->isDeletable())
        {
            delete pIn;
        }
    }

            
    if (out.size() == 0 || out.size() > 3)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d to %d expected.\n"), m_pCallFunctionName[EVENTS], 1,3);
        throw ast::InternalError(errorMsg);    
    }

    for (int i=0; i < out.size(); i++)
    {
        if (out[i]->isDouble() && out[i]->getAs<types::Double>()->isComplex() == false)
        {
            types::Double *pDbl = out[i]->getAs<types::Double>();
            m_iNbEvents = (i == 0) ? pDbl->getSize() : m_iNbEvents;
            if (pDbl->getSize() != m_iNbEvents)
            {
                sprintf(errorMsg, _("%s: Wrong size for output argument #%d: expecting %d.\n"), m_pCallFunctionName[EVENTS], i+1, m_iNbEvents);
                throw ast::InternalError(errorMsg);                
            }
            if (i == 1)
            {
                for (int k = 0; k<m_iNbEvents; k++)
                {
                    m_iVecEventIsTerminal.push_back(pDbl->get(k));
                }
            }
            else if (i == 2)
            {
                for (int k = 0; k<m_iNbEvents; k++)
                {
                    m_iVecEventDirection.push_back(pDbl->get(k));
                }
            }
        }
        else
        {
            sprintf(errorMsg, _("%s: Wrong type for output argument #%d: A real matrix expected.\n"), m_pCallFunctionName[EVENTS], i+1);
            throw ast::InternalError(errorMsg);          
        }
        m_iSizeOfInput[EVENTS] = m_iNbEvents;
        out[i]->killMe();
    }
}

