//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "odeparameters.hxx"

void OdeManager::init()
{
    char errorMsg[256];

    if (std::isnan(m_dblOptT0))
    {
        if (m_pDblTSpan->getSize() > 1)
        {
            m_dblT0 = m_pDblTSpan->get(0);   
        }
        else
        {
            sprintf(errorMsg, _("%ls: parameter \"%ls\" must be defined by user.\n"), m_wstrCaller.c_str(), L"t0");
            throw ast::InternalError(errorMsg);
        } 
    }
    else
    {
        m_dblT0 = m_dblOptT0;
    }

    // Complex ode is handled by using [real(y); imag(y)] as state, 
    // hence the number of *real* equations is twice the number of complex equatioms
    m_iNbRealEq = m_odeIsComplex ? 2*m_iNbEq : m_iNbEq;

    if (create() == NULL) 
    {
        sprintf(errorMsg,"Solver create error\n");
        throw ast::InternalError(errorMsg);
    }
    // Load Y0 into N_Serial vector
    // When ODE is complex  m_N_VectorY is the concatenation of real and imaginary part of user Y

    std::copy(m_pDblY0->get(), m_pDblY0->get()+m_iNbEq, NV_DATA_S(m_N_VectorY));
    if (m_odeIsComplex)
    {
        std::copy(m_pDblY0->getImg(), m_pDblY0->getImg()+m_iNbEq, NV_DATA_S(m_N_VectorY)+m_iNbEq);        
    }
    
    // Load Yp0 if DAE solver
    if (m_isDAE)
    {
        std::copy(m_pDblYp0->get(), m_pDblYp0->get()+m_iNbEq, NV_DATA_S(m_N_VectorYp));
        if (m_odeIsComplex)
        {
            std::copy(m_pDblYp0->getImg(), m_pDblYp0->getImg()+m_iNbEq, NV_DATA_S(m_N_VectorYp)+m_iNbEq);        
        }        
    }

    if (initialize(errorMsg))
    {
        throw ast::InternalError(errorMsg);
    }

    if (setUserData())
    {
        sprintf(errorMsg,"setUserData error\n");
        throw ast::InternalError(errorMsg);
    }

    if (m_dblInitialStep > 0 && setInitStep())
    {
        sprintf(errorMsg,"setInitStep error\n");
        throw ast::InternalError(errorMsg);
    }

    if (m_dblMaxStep > 0 && setMaxStep())
    {
        sprintf(errorMsg,"setMaxStep error\n");
        throw ast::InternalError(errorMsg);
    }

    if (m_dblMinStep > 0 && setMinStep())
    {
        sprintf(errorMsg,"setMinStep error\n");
        throw ast::InternalError(errorMsg);
    }

    // Events
    if (m_iNbEvents > 0 && setEventFunction())
    {
        sprintf(errorMsg,"setEventFunction error\n");
        throw ast::InternalError(errorMsg);        
    }

    // Absolute (a vector) and relative tolerance
    m_N_VectorAtol = N_VNew_Serial(m_iNbRealEq);
    if (m_odeIsComplex)
    {
        m_dblVecAtol.resize(m_iNbRealEq);
        for (int i=0; i<m_iNbEq; i++)
        {
            m_dblVecAtol[i+m_iNbEq] = m_dblVecAtol[i];
        }
    }
    std::copy(m_dblVecAtol.begin(), m_dblVecAtol.end(), NV_DATA_S(m_N_VectorAtol));
    if (setTolerances())
    {
        sprintf(errorMsg,"setTolerances error\n");
        throw ast::InternalError(errorMsg);
    }

    if (setMaxOrder())
    {
        sprintf(errorMsg,"setMaxOrder error\n");
        throw ast::InternalError(errorMsg);
    }

    if (setSolverAndJacobian(errorMsg))
    {
        throw ast::InternalError(errorMsg);        
    }

    if (computeIC(errorMsg))
    {
        throw ast::InternalError(errorMsg);        
    }

    if (setHandlerFunction())
    {
        sprintf(errorMsg,"setHandlerFunction error\n");
        throw ast::InternalError(errorMsg);
    }
    
}